﻿/*
    Photobooth a simple application to take pictures
    Copyright (C) 2024 Mike Nolan

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/
using System;
using System.Threading.Tasks;
using SDL2;
using SixLabors.ImageSharp;
using SixLabors.ImageSharp.PixelFormats;
using static SDL2.SDL;
using FlashCap;
using System.IO;
using System.Linq;
using System.Threading;
using System.Diagnostics;
namespace PhotoBooth
{
    public class PhotoBoothApp
    {
        Rgba32 Black = Color.Red.ToPixel<Rgba32>();

        private IntPtr IMG_LoadTexture(IntPtr renderer,byte[] buffer)
        {
            unsafe{
                fixed(byte* a = buffer)
                {
                    IntPtr ptr=(IntPtr)a;

                    IntPtr rw = SDL_RWFromConstMem(ptr,buffer.Length);

                    var img=SDL_image.IMG_Load_RW(rw,0);
                    
                    var res=SDL_CreateTextureFromSurface(renderer,img);
                    SDL_FreeSurface(img);
                    SDL_FreeRW(rw);
                    return res;
                }
            }
        }
        public async Task ImageProcessor(Func<byte[],Task> cb,int cam,int cap)
        {
                   var devices = new CaptureDevices();
           
            var descriptor0 = devices.EnumerateDescriptors().ElementAt(cam);

using(var device = await descriptor0.OpenAsync(descriptor0.Characteristics[cap]
    ,
    async bufferScope =>
    {
        // Captured into a pixel buffer from an argument.

        // Get image data (Maybe DIB/JPEG/PNG):
        byte[] image = bufferScope.Buffer.ExtractImage();
      
        await cb(image);
        // ...
    })){

// Start processing:
        await device.StartAsync();
            while (running)
            {
                
            }
        await device.StopAsync();
        }
        }
        bool running=true;
        public void Run(string[] args)
        {
            
            if(args.Length < 2)
            {
                var devices = new CaptureDevices();
                int i=0;
                Console.WriteLine("PhotoBooth Copyright (C) 2024 Mike Nolan");
                Console.WriteLine("Source code: https://gitlab.tesses.net/tesses50/PhotoBooth");
                
                Console.WriteLine("License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>");
                Console.WriteLine("This is free software; you are free to change and redistribute it.");
                Console.WriteLine("There is NO WARRANTY, to the extent permitted by law.");
                Console.WriteLine();
                Console.WriteLine("Press SPACE to take picture");
                Console.WriteLine("Run with arguments cameraNumber resolutionNumber from the []");
                Console.WriteLine("Run with additional argument for script, the first argument of script will be file path");
                Console.WriteLine();
                Console.WriteLine("Capture devices");
                foreach(var device in devices.EnumerateDescriptors())
                {
                    Console.WriteLine($"[{i++}] {device.Name} {device.Identity} {device.Description}");
                    int j=0;
                    foreach(var _cap in device.Characteristics)
                    {
                        Console.WriteLine($"\t[{j++}] {_cap.PixelFormat} {_cap.Width}x{_cap.Height} {_cap.Description}");
                    }
                }
                return;
            }
            

            if(!int.TryParse(args[0],out var cam))
                cam = 0;
            
            if(!int.TryParse(args[1],out var cap))
                cap = 0;

            
            SDL_image.IMG_Init(SDL_image.IMG_InitFlags.IMG_INIT_JPG|SDL_image.IMG_InitFlags.IMG_INIT_PNG| SDL_image.IMG_InitFlags.IMG_INIT_TIF| SDL_image.IMG_InitFlags.IMG_INIT_WEBP);

            var window=SDL_CreateWindow("Photo Booth",SDL_WINDOWPOS_CENTERED,SDL_WINDOWPOS_CENTERED,640,480,SDL_WindowFlags.SDL_WINDOW_RESIZABLE);
            var renderer = SDL_CreateRenderer(window,-1,SDL_RendererFlags.SDL_RENDERER_SOFTWARE);
            SemaphoreSlim slim=new SemaphoreSlim(1,1);
            byte[] pic=new byte[0];

            Task.Run(async()=>await ImageProcessor(async e=>{
                await slim.WaitAsync();
                pic = e;
                slim.Release();
            },cam,cap)).Wait(0);

            while(running)
            {
                 
                slim.Wait();
                var image = pic;
                slim.Release();
                
        // Anything use of it...
           if(SDL_PollEvent(out var _event) > 0)
                {   
                    switch(_event.type){
                        case SDL_EventType.SDL_QUIT:
                            running=false;
                            break;
                        case SDL_EventType.SDL_KEYDOWN:
                            if(_event.key.keysym.sym == SDL_Keycode.SDLK_SPACE)
                            {
                                string filename = $"{DateTime.Now.ToString("yyyyMMdd_HHmmss")}.jpg";
                                File.WriteAllBytes(filename,image);
                                Console.WriteLine($"Captured to: {Path.GetFullPath(filename)}");
                                if(args.Length == 3)
                                {
                                    Process.Start(args[2],$"\"{filename}\"");
                                }
                            }
                            break;
                    } 
                }
       
        if(image.Length == 0) continue;
        var tex = IMG_LoadTexture(renderer,image);
       

        

        SDL_QueryTexture(tex,out var fmt,out var acc,out var w,out var h);


        SDL_GetWindowSize(window,out var wW,out var wH);



                SDL_Rect src = new SDL_Rect
                {
                    w = w,
                    h = h,
                    x = 0,
                    y = 0
                };

                SDL_Rect dest = new SDL_Rect
                {
                    x = 0,
                    y = 0,
                    w = wW,
                    h = wH
                };
               
        SDL_SetRenderDrawColor(renderer,255,0,0,255);
        SDL_RenderClear(renderer);

     

        SDL_RenderCopy(renderer,tex,ref src,ref dest);

        

        
        SDL_RenderPresent(renderer);
        SDL_DestroyTexture(tex);
            
            
            }
            SDL_DestroyRenderer(renderer);
            SDL_DestroyWindow(window);
        }
    }
}
